@@ -112,7 +112,7 @@ has sigfile  => (is => 'rw');
 # 2 - Line options read/write attributes
 
 foreach (qw(
-    component hrefdecode repacksuffix unzipopt searchmode
+    component ctype hrefdecode repacksuffix unzipopt searchmode
     dirversionmangle downloadurlmangle dversionmangle filenamemangle pagemangle
     oversionmangle oversionmanglepagemangle pgpsigurlmangle uversionmangle
     versionmangle
@@ -167,6 +167,10 @@ has decompress => (
     is      => 'rw',
     default => sub { 0 },
 );
+has gitexport => (
+    is      => 'rw',
+    default => sub { 'default' },
+);
 has gitmode => (
     is      => 'rw',
     default => sub { 'shallow' },
@@ -466,7 +470,9 @@ EOF
                 #
                 # $ regexp-assemble <<EOF
                 # component
+                # ctype
                 # date
+                # gitexport
                 # gitmode
                 # hrefdecode
                 # mode
@@ -477,7 +483,7 @@ EOF
                 # unzipopt
                 # EOF
                 elsif ($opt
-                    =~ /^\s*((?:(?:(?:search|git)?m|hrefdec)od|dat)e|(?:componen|unzipop)t|p(?:gpmode|retty)|repacksuffix)\s*=\s*(.+?)\s*$/
+                    =~ /^\s*((?:(?:(?:(?:search)?m|hrefdec)od|dat)e|c(?:omponent|type)|git(?:export|mode)|p(?:gpmode|retty)|repacksuffix|unzipopt))\s*=\s*(.+?)\s*$/
                 ) {
                     $self->$1($2);
                 } elsif ($opt =~ /^\s*versionmangle\s*=\s*(.+?)\s*$/) {
@@ -611,6 +617,12 @@ EOF
               . "  $self->{line}";
             return $self->status(1);
         }
+        if ($self->mode ne 'git' and $self->gitexport ne 'default') {
+            uscan_warn "gitexport option is valid only in git mode,\n"
+              . "  ignoring gitexport in $watchfile:\n"
+              . "  $self->{line}";
+            return $self->status(1);
+        }
 
         # Limit use of opts="repacksuffix" to the single upstream package
         if ($self->repacksuffix and @{ $self->shared->{components} }) {
@@ -735,6 +747,24 @@ EOF
 
     }
 
+    if ($self->ctype) {
+        my $version;
+        my $mod = "Devscripts::Uscan::Ctype::$self->{ctype}";
+        eval "require $mod";
+        if ($@) {
+            uscan_warn "unknown ctype $self->{ctype}";
+            uscan_debug $@;
+            return $self->status(1);
+        }
+        my $dir            = $self->component || '.';
+        my $ctypeTransform = $mod->new({ dir => $dir });
+        if ($version = $ctypeTransform->version) {
+            $lastversion = $version;
+            uscan_verbose "Found version $version for component $dir";
+            $self->versionmode('newer');
+        }
+    }
+
     # End parsing the watch line for all version=1/2/3/4
     # all options('...') variables have been set
 
@@ -817,7 +847,8 @@ EOF
 "Unable to set versionmode=prev for the line without opts=pgpmode=prev\n"
                   . "  in $watchfile, skipping:\n  $self->{line}";
             } else {
-                uscan_warn "Nothing was downloaded before, skipping pgp check";
+                uscan_verbose
+                  "Nothing was downloaded before, skipping pgp check";
                 uscan_verbose "  line " . $self->line;
             }
             return $self->status(1);
@@ -854,7 +885,7 @@ EOF
 
             # Find the path with the greatest version number matching the regex
             $base
-              = recursive_regex_dir($self->downloader, $base,
+              = recursive_regex_dir($self, $base,
                 $self->dirversionmangle, $watchfile, \$self->line,
                 $self->shared->{download_version});
             if ($base eq '') {
@@ -866,6 +897,7 @@ EOF
             # It's fine even for ftp sites
             $basedir = $base;
             $basedir =~ s%^\w+://[^/]+/%/%;
+            $basedir =~ s%/[^/]*(?:[#?].*)?$%/%;
             $pattern
               = "(?:(?:$site)?" . quotemeta($basedir) . ")?$filepattern";
         } else {
@@ -1116,8 +1148,8 @@ sub cmp_versions {
 
             # There's a newer upstream version available, which may already
             # be on our system or may not be
-            uscan_msg "   => Newer package available from\n"
-              . "      $self->{upstream_url}";
+            uscan_msg " => Newer package available from:\n"
+              . "        => $self->{upstream_url}";
             $dehs_tags->{'status'} //= "newer package available";
             $main::found++;
         } elsif ($compver eq 'same') {
@@ -1129,13 +1161,13 @@ sub cmp_versions {
                 ? ""
                 : " (mangled local version is $mangled_lastversion)\n"
               );
-            uscan_verbose "   => Package is up to date for from\n"
-              . "      $self->{upstream_url}";
+            uscan_verbose " => Package is up to date from:\n"
+              . "             => $self->{upstream_url}";
             $dehs_tags->{'status'} //= "up to date";
             if ($self->shared->{download} > 1) {
 
                 # 2=force-download or 3=overwrite-download
-                uscan_verbose "   => Forcing download as requested";
+                uscan_verbose " => Forcing download as requested";
                 $main::found++;
             } else {
                 # 0=no-download or 1=download
@@ -1150,11 +1182,11 @@ sub cmp_versions {
                 ? ""
                 : " (mangled local version is $mangled_lastversion)\n"
               );
-            uscan_verbose "   => Only older package available from\n"
-              . "      $self->{upstream_url}";
+            uscan_verbose " => Only older package available from:\n"
+              . "             => $self->{upstream_url}";
             $dehs_tags->{'status'} //= "only older package available";
             if ($self->shared->{download} > 1) {
-                uscan_verbose "   => Forcing download as requested";
+                uscan_verbose " => Forcing download as requested";
                 $main::found++;
             } else {
                 $self->shared->{download} = 0;
@@ -1180,6 +1212,8 @@ Download file and, if available and needed, signature files.
 
 =cut
 
+my %already_downloaded;
+
 sub download_file_and_sig {
     my ($self) = @_;
     uscan_debug "line: download_file_and_sig()";
@@ -1191,11 +1225,20 @@ sub download_file_and_sig {
         return 0;
     }
 
+    # configure downloader
+    $self->downloader->git_export_all($self->gitexport eq 'all');
+
     # 6.1 download tarball
     my $download_available = 0;
+    my $upstream_base      = basename($self->upstream_url);
     $self->signature_available(0);
     my $sigfile;
     my $sigfile_base = $self->newfile_base;
+    uscan_die
+"Already downloaded a file named $self->{newfile_base}: use filenamemangle to avoid this"
+      if ($already_downloaded{ $self->{newfile_base} });
+    $already_downloaded{ $self->{newfile_base} } = 1;
+
     if ($self->pgpmode ne 'previous') {
 
         # try download package
@@ -1217,7 +1260,7 @@ sub download_file_and_sig {
                   "Successfully downloaded package: $self->{newfile_base}\n";
             } else {
                 dehs_verbose
-"Failed to download upstream package: $self->{newfile_base}\n";
+                  "Failed to download upstream package: $upstream_base\n";
             }
         } elsif (-e "$self->{config}->{destdir}/$self->{newfile_base}") {
             $download_available = 1;
@@ -1225,8 +1268,7 @@ sub download_file_and_sig {
               "Not downloading, using existing file: $self->{newfile_base}\n";
             $skip_git_vrfy = 1;
         } elsif ($self->shared->{download} > 0) {
-            uscan_verbose
-              "Downloading upstream package: $self->{newfile_base}";
+            uscan_verbose "Downloading upstream package: $upstream_base";
             $download_available = $self->downloader->download(
                 $self->upstream_url,
                 "$self->{config}->{destdir}/$self->{newfile_base}",
@@ -1238,15 +1280,18 @@ sub download_file_and_sig {
             );
             if ($download_available) {
                 dehs_verbose
-                  "Successfully downloaded package: $self->{newfile_base}\n";
+                  "Successfully downloaded upstream package: $upstream_base\n";
+                if (@{ $self->filenamemangle }) {
+                    dehs_verbose
+                      "Renamed upstream package to: $self->{newfile_base}\n";
+                }
             } else {
                 dehs_verbose
-"Failed to download upstream package: $self->{newfile_base}\n";
+                  "Failed to download upstream package: $upstream_base\n";
             }
         } else {    # $download = 0,
             $download_available = 0;
-            dehs_verbose
-              "Not downloading upstream package: $self->{newfile_base}\n";
+            dehs_verbose "Not downloading upstream package: $upstream_base\n";
         }
     }
     if ($self->pgpmode eq 'self') {
@@ -1279,7 +1324,7 @@ sub download_file_and_sig {
         # Decompress archive if requested and applicable
         if ($download_available == 1 and $self->{'decompress'}) {
             my $suffix_gz = $sigfile_base;
-            $suffix_gz =~ s/.*?(\.gz|\.xz|\.bz2|\.lzma)?$/$1/;
+            $suffix_gz =~ s/.*?(\.gz|\.xz|\.bz2|\.lzma|\.zstd?)?$/$1/;
             if ($suffix_gz eq '.gz') {
                 if (-x '/bin/gunzip') {
                     uscan_exec('/bin/gunzip', "--keep",
@@ -1316,6 +1361,15 @@ sub download_file_and_sig {
                     uscan_warn "Please install xz-utils or lzma.";
                     return $self->status(1);
                 }
+            } elsif ($suffix_gz =~ /^zstd?$/) {
+                if (-x '/usr/bin/unzstd') {
+                    uscan_exec('/usr/bin/unzstd', "--keep",
+                        "$self->{config}->{destdir}/$sigfile_base");
+                    $sigfile_base =~ s/(.*?)\.zst/$1/;
+                } else {
+                    uscan_warn("Please install zstd.\n");
+                    return $self->status(1);
+                }
             } else {
                 uscan_die "Unknown type file to decompress: $sigfile_base";
             }
@@ -1376,7 +1430,7 @@ sub download_file_and_sig {
         }
         $sigfile = "$sigfile_base.$suffix_sig";
         if ($self->shared->{signature} == 1) {
-            uscan_verbose "Downloading OpenPGP signature from\n"
+            uscan_verbose "Downloading OpenPGP signature from:\n"
               . "   $pgpsig_url (pgpsigurlmangled)\n   as $sigfile";
             $self->signature_available(
                 $self->downloader->download(
@@ -1386,7 +1440,7 @@ sub download_file_and_sig {
                     $self->mode
                 ));
         } else {    # -1, 0
-            uscan_verbose "Not downloading OpenPGP signature from\n"
+            uscan_verbose "Not downloading OpenPGP signature from:\n"
               . "   $pgpsig_url (pgpsigurlmangled)\n   as $sigfile";
             $self->signature_available(
                 (-e "$self->{config}->{destdir}/$sigfile") ? 1 : 0);
@@ -1395,7 +1449,7 @@ sub download_file_and_sig {
         $pgpsig_url = $self->upstream_url;
         $sigfile    = $self->newfile_base;
         if ($self->shared->{signature} == 1) {
-            uscan_verbose "Downloading OpenPGP signature from\n"
+            uscan_verbose "Downloading OpenPGP signature from:\n"
               . "   $pgpsig_url (pgpmode=previous)\n   as $sigfile";
             $self->signature_available(
                 $self->downloader->download(
@@ -1405,7 +1459,7 @@ sub download_file_and_sig {
                     $self->mode
                 ));
         } else {    # -1, 0
-            uscan_verbose "Not downloading OpenPGP signature from\n"
+            uscan_verbose "Not downloading OpenPGP signature from:\n"
               . "   $pgpsig_url (pgpmode=previous)\n   as $sigfile";
             $self->signature_available(
                 (-e "$self->{config}->{destdir}/$sigfile") ? 1 : 0);
@@ -1481,8 +1535,12 @@ sub download_file_and_sig {
             return $self->status(1);
         } else {
             my ($gitrepo, $gitref) = split /[[:space:]]+/, $self->upstream_url;
-            $self->keyring->verify_git($self->pkg . "-temporary.$$.git",
-                $gitref, $self->downloader->git_upstream);
+            $self->keyring->verify_git(
+                "$self->{downloader}->{destdir}/"
+                  . $self->pkg
+                  . "-temporary.$$.git",
+                $gitref, $self->downloader->git_upstream
+            );
         }
     } else {
         uscan_warn "strange ... unknown pgpmode = $self->{pgpmode}";
@@ -1556,7 +1614,7 @@ sub mkorigtargz {
         if ($Devscripts::MkOrigtargz::found_comp) {
             uscan_verbose
               "Forcing compression to $Devscripts::MkOrigtargz::found_comp";
-            $self->force_repack(1);
+            $self->repack(1);
         }
         @ARGV = ();
         push @ARGV, "--package", $self->pkg;
